import os, re, time
import flopy
import sys
import numpy as np
import pyvista as pv
from scipy.interpolate import griddata

class Mf6VtkGenerator:
    def __init__(self, sim, vtkDir):
        self.sim = sim
        self.vtkDir = vtkDir
    
    def listModels(self):
        print("\n Models in simulation: %s"%self.sim.model_names)
        
    def loadModel(self, modelName):
        self.gwf = self.sim.get_model(modelName)
        self.packageList = self.gwf.get_package_list()
        print("Package list: %s"%self.packageList)
    
    # #utils
    # def deleteAllFiles(self):
    #     try:
    #         # Check if the folder exists
    #         if not os.path.exists(self.vtkDir):
    #             os.mkdir(self.vtkDir)
    #             print(f"Folder created: {file_path}")

    #         # Iterate through all files in the folder and delete them
    #         print(f"Check for files in '{self.vtkDir}'.")
    #         for filename in os.listdir(self.vtkDir):
    #             file_path = os.path.join(self.vtkDir, filename)
    #             if os.path.isfile(file_path):
    #                 os.remove(file_path)
    #                 print(f"Deleted file: {file_path}")

    #         print(f"\nAll files in '{self.vtkDir}' have been deleted.\n")

    #     except Exception as e:
    #         print(f"An error occurred: {str(e)}")
    
    def saveArray(self,outputDir, array, arrayName):
        if array is not None:
            np.save(os.path.join(outputDir,'param_'+arrayName),array)    
            
    def timing_decorator(func):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = func(*args, **kwargs)
            end_time = time.time()
            execution_time = end_time - start_time
            print(f"Vtk file took {execution_time:.4f} seconds to be generated.")
            return result
        return wrapper
    
    @timing_decorator
    def exportBc(self,bcon,nper):
        #open package and filter names
        bcObj = self.gwf.get_package(bcon)  
        bcObjSpdNames = bcObj.stress_period_data.dtype.names[1:]
        print('Working for %s package, creating the datasets: %s'%(bcon,bcObjSpdNames))
        #create a flat index
        flatIndexList = []
        #get the first stress period that has the active bc
        bcKeys = list(bcObj.stress_period_data.data.keys())
        for row in bcObj.stress_period_data.data[bcKeys[nper]]:
            # print(row.cellid)
            # print(self.gwf.modelgrid.shape)
            flatIndex = np.ravel_multi_index(row.cellid,self.gwf.modelgrid.shape)
            flatIndexList.append(flatIndex)
        #empty object
        
        #loop over the cells
        if not re.search('rch',bcon,re.IGNORECASE):
            cropVtk = pv.UnstructuredGrid()
            for index, cell in enumerate(flatIndexList):
                tempVtk = self.vtkGeom.extract_cells(cell)
                for name in bcObjSpdNames:
                    tempVtk.cell_data[name] = bcObj.stress_period_data.data[nper][name][index]
                cropVtk += tempVtk

                if index % 500 == 0:
                    print('... %d cells generated'%index)
                    #print(index, name, nper, bcObj.stress_period_data.data[nper][name][index])
        else:
            cropVtk = self.vtkGeom.extract_cells(flatIndexList)
            for name in bcObjSpdNames:
                cropVtk.cell_data[name] = bcObj.stress_period_data.data[nper][name]
                
        #save and return
        cropVtk.save(os.path.join(self.vtkDir,bcon+'.vtk'))
        #return cropVtk
        
    @timing_decorator
    def exportObs(self,obs):
        #open package 
        obsObj = self.gwf.get_package(obs)
        obsKey = list(obsObj.continuous.data.keys())[0]
        obsObjNames = obsObj.continuous.data[obsKey].dtype.names[:2]
        bcObjArray = obsObj.continuous.data[obsKey].id
        print(bcObjArray)
        #create a flat index
        flatIndexList = []
        for cell in bcObjArray:
            flatIndex = np.ravel_multi_index(cell,self.gwf.modelgrid.shape)
            flatIndexList.append(flatIndex)
        #empty object
        cropVtk = pv.UnstructuredGrid()
        #loop over the cells
        for index, cell in enumerate(flatIndexList):
            tempVtk = self.vtkGeom.extract_cells(cell)
            for name in obsObjNames:
                tempVtk.cell_data[name] = np.array([obsObj.continuous.data[obsKey][name][index]])
            cropVtk += tempVtk
        #save and return
        cropVtk.save('../Vtk/%s.vtk'%obs)
        #return cropVtk
        
    def generateGeometryArrays(self):
        dis = self.gwf.get_package('DIS')
        dis.export(self.vtkDir, fmt='vtk', smooth=True, binary=False)
        
        if self.gwf.get_package('DIS') is not None:
            dis = self.gwf.get_package('DIS')
            idomainArray = dis.idomain.array
            self.saveArray(self.vtkDir, idomainArray, 'idomain')

        if self.gwf.get_package('IC') is not None:
            ic = self.gwf.get_package('IC')
            strtArray = ic.strt.array
            self.saveArray(self.vtkDir, strtArray, 'strt')

        if self.gwf.get_package('STO') is not None:
            sto = self.gwf.get_package('STO')
            iconvertArray = sto.iconvert.array
            syArray = sto.sy.array
            ssArray = sto.ss.array
            self.saveArray(self.vtkDir, iconvertArray, 'iconvert')
            self.saveArray(self.vtkDir, syArray, 'sy')
            self.saveArray(self.vtkDir, ssArray, 'ss')

        if self.gwf.get_package('NPF') is not None:
            npf = self.gwf.get_package('NPF')
            icelltypeArray = npf.icelltype.array
            kArray = npf.k.array
            k22Array = npf.k22.array
            k33Array = npf.k33.array
            wetdryArray = npf.wetdry.array
            self.saveArray(self.vtkDir, icelltypeArray, 'icelltype')
            self.saveArray(self.vtkDir, kArray, 'k')
            self.saveArray(self.vtkDir, k22Array, 'k22')
            self.saveArray(self.vtkDir, k33Array, 'k33')
            self.saveArray(self.vtkDir, wetdryArray, 'wetdry')
        
        # Build geometry vtk
        if self.gwf.modelgrid.grid_type == 'structured':
            vtkFile = os.path.join(self.vtkDir,"dis.vtk")
        elif self.gwf.modelgrid.grid_type == 'vertex':
            vtkFile = os.path.join(self.vtkDir,"disv.vtk")
        elif self.gwf.modelgrid.grid_type == 'unstructured':
            vtkFile = os.path.join(self.vtkDir,"disu.vtk")
        else:
            print("No Dis file was found")
        gwfGrid = pv.read(vtkFile)
        gwfGrid.clear_cell_data()
        #self.geomPath = ('../Model/Vtk/modelGeometry.vtk')
        self.geomPath = os.path.join(self.vtkDir,'modelGeometry.vtk')
        gwfGrid.save(self.geomPath)
        os.remove(vtkFile)
        self.vtkGeom = gwfGrid
        
    def generateParamVtk(self):
        vtkParam = pv.read(self.geomPath)
        paramList = [file for file in os.listdir(self.vtkDir) if file.startswith('param')]
        
        for param in paramList:
            paramName = re.split('[_;.]',param)[1]
            paramValues = np.load(os.path.join(self.vtkDir,param),allow_pickle=True)
            vtkParam.cell_data[paramName] = paramValues.flatten()
        paramPath = os.path.join(self.vtkDir,'modelParameters.vtk')
        vtkParam.save(paramPath)
        print("Parameter Vtk Generated")
        
    def generateBcObsVtk(self, nper):
        vtkGrid = pv.read(self.geomPath)
        bcList = [x for x in self.packageList if re.search(r'\d',x) and not re.search('obs',x,re.IGNORECASE)]
        obsList = [x for x in self.packageList if re.fullmatch('obs',x,re.IGNORECASE)]
        #print(bcList)
        for bc in bcList:
            if bc[3]!='A':
                self.exportBc(bc, nper)
                print("%s vtk generated"%bc)
        for obs in obsList:
            self.exportObs(obs)
            print("%s btk generated"%obs)

    def generateHeadVtk(self, nper, crop=False):
        headObj = self.gwf.output.head()
        heads = headObj.get_data(kstpkper=(0,nper))
        waterTable = flopy.utils.postprocessing.get_water_table(heads)

        geomVtk = pv.read(self.geomPath)
        geomVtk.clear_cell_data()
        geomVtk.cell_data['heads'] = heads.flatten()

        if crop:
            botmArray = self.gwf.modelgrid.botm.flatten()
            #wtArray = np.repeat(waterTable,self.gwf.modelgrid.nlay).filled()
            wtArray = np.hstack([waterTable for i in range(self.gwf.modelgrid.nlay)]) 
            activeArray = np.where(wtArray > botmArray, 1, 0)
            geomVtk.cell_data['active'] = activeArray

        # geomVtk = geomVtk.threshold(value=(1,1), scalars="heads")
        # geomVtk = geomVtk.cell_data_to_point_data()
        geomVtk.save(os.path.join(self.vtkDir,'waterHeads.vtk'))

    def generateWaterTableVtk(self, nper):
        headObj = self.gwf.output.head()
        heads = headObj.get_data(kstpkper=(0,nper))
        waterTable = flopy.utils.postprocessing.get_water_table(heads)

        xCell = self.gwf.modelgrid.xcellcenters
        yCell = self.gwf.modelgrid.ycellcenters
        zCell = waterTable.data

        xyVerts = self.gwf.modelgrid.verts

        # Interpolate z-values at new xy locations
        zVert = griddata((xCell, yCell), zCell, xyVerts, method='nearest')
        
        faces = np.hstack([cell[3:] for cell in self.gwf.modelgrid.cell2d])
        points = list(zip(self.gwf.modelgrid.verts[:,0],self.gwf.modelgrid.verts[:,1],zVert))
        mesh = pv.PolyData(points, faces)

        mesh.cell_data['waterTable'] = waterTable

        mesh = mesh.cell_data_to_point_data()
        mesh.save(os.path.join(self.vtkDir,'waterTable.vtk'))



import rasterio
import geopandas as gpd
from shapely.geometry import Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon

def getLayCellElevTupleFromRaster(gwf,interIx,rasterPath,geomPath):
    rasterSrc = rasterio.open(rasterPath)
    geomSrc = gpd.read_file(geomPath)
    insideCellsIds = []
    layCellTupleList = []
    cellElevList = []

    #model parameters
    nlay = gwf.modelgrid.nlay
    xCenter = gwf.modelgrid.xcellcenters
    yCenter = gwf.modelgrid.ycellcenters
    rasterElev = [elev[0] for elev in rasterSrc.sample(zip(xCenter,yCenter))] 
    topBotm = gwf.modelgrid.top_botm

    #working with the cell ids
    #loop over the geometries to get the cellids
    for index, row in geomSrc.iterrows():
        tempCellIds = interIx.intersect(row.geometry).cellids
        for cell in tempCellIds:
            insideCellsIds.append(cell)

    #working with the cell elevations and create laycell tuples
    for cell in insideCellsIds:
        #looping over elevations
        if topBotm[-1, cell] < rasterElev[cell] <= topBotm[0,cell]:
            cellElevList.append(rasterElev[cell])
        else: 
            print('The cell %d has a elevation of %.2f outside the model vertical domain'%(cell,rasterElev[cell]))
        #looping through layers
        for lay in range(nlay):  
            if topBotm[lay+1, cell] < rasterElev[cell] <= topBotm[lay,cell]:
                layCellTupleList.append((lay,cell))

    return layCellTupleList, cellElevList

def getLayCellElevTupleFromElev(gwf,interIx,elevValue,geomPath):
    geomSrc = gpd.read_file(geomPath)
    insideCellsIds = []
    layCellTupleList = []

    #model parameters
    nlay = gwf.modelgrid.nlay
    topBotm = gwf.modelgrid.top_botm

    #working with the cell ids
    #loop over the geometries to get the cellids
    for index, row in geomSrc.iterrows():
        tempCellIds = interIx.intersect(row.geometry).cellids
        for cell in tempCellIds:
            insideCellsIds.append(cell)

    #working with the cell elevations
    for cell in insideCellsIds:
        for lay in range(nlay):  # Loop through layers\n",
            if topBotm[lay+1, cell] < elevValue <= topBotm[lay,cell]:
                layCellTupleList.append((lay,cell))


    return layCellTupleList
        
    






import os, io, fiona, folium
import geopandas as gpd
from scipy.spatial import Voronoi
import numpy as np
import matplotlib.pyplot as plt
from shapely.geometry import Point
#from .utils import add_to_map
#from mf6.z3graphicsapp.utils import heads_to_map

# from .models import (LimitLayer,
#     RefinementLayerNumber,
#     RefinementLayer,
#     VoronoiMesh,
#     Cell2d,
#     VerticalLayerSetup,
#     VerticalLayerByRaster,
#     ModelArray)

# from colour import Color

# Select a colormap
cmap = plt.cm.viridis

def read_shp_from_zip(file):
    zipshp = io.BytesIO(open(file, 'rb').read())
    with fiona.BytesCollection(zipshp.read()) as src:
        crs = src.crs
        gdf = gpd.GeoDataFrame.from_features(src, crs=crs)
    return gdf
         
def plotOrgDistPoints(vorMesh):
  limitXY = vorMesh.modelDis['limitGeometry']
  limitDf = gpd.GeoDataFrame([{'geometry':limitXY}])

  orgPoints = np.array(vorMesh.modelDis['vertexOrg'])

  tempDistPoints = []
  for key, value in vorMesh.modelDis['vertexDist'].items()  :
     tempDistPoints += value
  distPoints = np.array(tempDistPoints)
  #print(distPoints)

  fig, ax = plt.subplots(figsize=(36,24))
  ax.scatter(orgPoints[:,0], orgPoints[:,1], 
             label='Original', alpha=0.5, ec='crimson', fc='none')
  ax.scatter(distPoints[:,0], distPoints[:,1], s=2, marker='^',
             label='Distributed', alpha=0.5, ec='slateblue')
  limitDf.plot(ax=ax, label='Limit', alpha=0.5, ec='teal', fc='none', ls='-')
  for key, layer in vorMesh.discLayers.items():
     distLayersList = []
     for layerGeom in layer['layerGeoms']:
        distLayersList.append(layerGeom)
     layerDf = gpd.GeoDataFrame(geometry=distLayersList)
     #layerDf.plot(ax=ax, alpha=0.5, label=key.split('_')[0])
     layerDf.plot(ax=ax, alpha=0.5, label=None)
  ax.legend(loc='upper left')

  plt.show()

def plotCirclesPoints(vorMesh):
  #limitXY = vorMesh.modelDis['limitGeometry']
  circleUnionDf = gpd.GeoDataFrame([{'geometry':vorMesh.modelDis['circleUnion']}])
  polyPointList = vorMesh.modelDis['vertexBuffer']
  orgPoints = np.array(vorMesh.modelDis['vertexOrg'])

  tempDistPoints = []
  for key, value in vorMesh.modelDis['vertexDist'].items():
     tempDistPoints += value
  distPoints = np.array(tempDistPoints)
  #distPoints = np.array(vorMesh.modelDis['vertexDist'])
  polyPoints = np.array(polyPointList)

  fig, ax = plt.subplots(figsize=(36,24))
  ax.scatter(orgPoints[:,0], orgPoints[:,1], 
             label='Original', alpha=0.5, ec='crimson', fc='none')
  ax.scatter(distPoints[:,0], distPoints[:,1], s=5, marker='^',
             label='Distributed', alpha=0.5, ec='slateblue')
  ax.scatter(polyPoints[:,0], polyPoints[:,1], s=5, marker='^',
            label='CirclePoints', alpha=0.5, ec='tan')
  circleUnionDf.plot(ax=ax, label='circleUnion', alpha=0.5, ec='teal', fc='none', ls='-')
  #polyPointDf.plot(ax=ax, label='Limit', alpha=0.5, ec='tan', fc='none', ls='-')

  for key, layer in vorMesh.discLayers.items():
    distLayersList = []
    for layerGeom in layer['layerGeoms']:
      distLayersList.append(layerGeom)
    layerDf = gpd.GeoDataFrame(geometry=distLayersList)
    layerDf.plot(ax=ax, alpha=0.5, label=None)
  ax.legend(loc='upper left')

  plt.show()

def plotKeyList(vorMesh, keyList):
  fig, ax = plt.subplots(figsize=(36,24))
  for key in keyList:
    keyDf = gpd.GeoDataFrame([{'geometry':vorMesh.modelDis[key]}])
    keyDf.plot(ax=ax, label=key, alpha=0.5, ec='teal', fc='none', ls='-')
  
  ax.legend(loc='upper left')

  plt.show()
     
def bcGraph(self, ref_obj, ref_type, *args, **kwargs):
    project=self.getProject(self.request)

    limitLayer= LimitLayer.objects.get(project=project)
    meshLayer= VoronoiMesh.objects.get(project=project)

    gdfLimit=read_shp_from_zip(limitLayer.limitFile.path)
    gdfMesh=read_shp_from_zip(meshLayer.meshFile.path)

    if ref_type == "Recharge":
        gdfRef=read_shp_from_zip(ref_obj.rchFile.path)
    elif ref_type == "Evapotranspiration":
        gdfRef=read_shp_from_zip(ref_obj.evtFile.path)
    else:
        gdfRef=read_shp_from_zip(ref_obj.refinementFile.path)

    gdfRef['id'] = np.arange(0,gdfRef.shape[0])

    meshmap = folium.Map(max_zoom=20)

    gridGroup = folium.FeatureGroup(name='Grid')
    limGroup = folium.FeatureGroup(name='Limit Layer')
    refGroup = folium.FeatureGroup(name=ref_type)

    basemap2 = ('https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}')
    folium.WmsTileLayer(
      url=basemap2,
      layers=None,
      name='Google Sattelite',
      attr='Google',
      control=True,
      overlay=False,
    ).add_to(meshmap)

    gridStyle = {'fillColor': '#5AAACD',"weight": 1,"opacity": 0.65,"fillOpacity": 0}
    gridLayer = folium.GeoJson(data=gdfMesh["geometry"], style_function = lambda x: gridStyle)
    gridGroup.add_child(gridLayer)

    if gdfRef["geometry"][0].geom_type == 'Point':
        refLayer = folium.GeoJson(data=gdfRef,popup=folium.GeoJsonPopup(fields=['id']),
        marker = folium.CircleMarker(radius = 1, # Radius in metres
                                       weight = 6, #outline weight
                                       color = '#7B7C7E',
                                       fill_opacity = 1))
        refGroup.add_child(refLayer)
    else:
        refStyle = {'color': '#00CED1',"weight": 4,"opacity": 0.65,"fillOpacity": 0.2,"radius":3}
        refLayer = folium.GeoJson(data=gdfRef, style_function = lambda x: refStyle,
            popup=folium.GeoJsonPopup(fields=['id']))
        refGroup.add_child(refLayer)

    limStyle = {'color': '#ef4310',"weight": 4,"opacity": 0.65,"fillOpacity": 0,"radius":3}
    limLayer = folium.GeoJson(data=gdfLimit["geometry"], style_function = lambda x: limStyle)
    limGroup.add_child(limLayer)

    meshmap.add_child(limGroup)
    meshmap.add_child(gridGroup)
    meshmap.add_child(refGroup)

    bPts = gdfMesh.to_crs("EPSG:4326").total_bounds
    xMinyMin = list(bPts[:2])[::-1]
    xMaxyMax = list(bPts[2:])[::-1]
    meshmap.fit_bounds([xMinyMin,xMaxyMax], padding=(1, 1))
    folium.LayerControl('topleft', collapsed= False).add_to(meshmap)

    meshmap = meshmap._repr_html_()

    return meshmap

def limitGraph(self, *args, **kwargs):
    project=self.getProject(self.request)
    limitLayer= LimitLayer.objects.get(project=project)

    gdfLimit=read_shp_from_zip(limitLayer.limitFile.path)

    meshmap = folium.Map(max_zoom=20)

    limGroup = folium.FeatureGroup(name='Limit Layer')

    basemap2 = ('https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}')
    folium.WmsTileLayer(
      url=basemap2,
      layers=None,
      name='Google Sattelite',
      attr='Google',
      control=True,
      overlay=False,
    ).add_to(meshmap)

    limStyle = {'color': '#ef4310',"weight": 4,"opacity": 0.65,"fillOpacity": 0,"radius":3}
    limLayer = folium.GeoJson(data=gdfLimit["geometry"], style_function = lambda x: limStyle)
    limGroup.add_child(limLayer)

    meshmap.add_child(limGroup)

    bPts = gdfLimit.to_crs("EPSG:4326").total_bounds
    xMinyMin = list(bPts[:2])[::-1]
    xMaxyMax = list(bPts[2:])[::-1]
    meshmap.fit_bounds([xMinyMin,xMaxyMax], padding=(1, 1))
    folium.LayerControl('topleft', collapsed= False).add_to(meshmap)

    meshmap = meshmap._repr_html_()

    return meshmap

def meshGraph(self,mesh_obj, *args, **kwargs):
    project=self.getProject(self.request)
    limitLayer= LimitLayer.objects.get(project=project)
    refinementLayers=RefinementLayer.objects.filter(project=project)

    gdfLimit=read_shp_from_zip(limitLayer.limitFile.path)
    gdfRefDict = {}
    for index, refinemet in enumerate(refinementLayers):
        gdfRefDict[index] = read_shp_from_zip(refinementLayers[index].refinementFile.path)
    gdfMesh = read_shp_from_zip(mesh_obj.meshFile.path)

    meshmap = folium.Map(max_zoom=20)

    gridGroup = folium.FeatureGroup(name='Grid')
    limGroup = folium.FeatureGroup(name='Limit Layer')
    refGroup = folium.FeatureGroup(name='Refinement Layers')

    basemap2 = ('https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}')
    folium.WmsTileLayer(
      url=basemap2,
      layers=None,
      name='Google Sattelite',
      attr='Google',
      control=True,
      overlay=False,
    ).add_to(meshmap)

    gridStyle = {'fillColor': '#5AAACD',"weight": 1,"opacity": 0.65,"fillOpacity": 0}
    gridLayer = folium.GeoJson(data=gdfMesh["geometry"], style_function = lambda x: gridStyle)
    gridGroup.add_child(gridLayer)

    meshmap.add_child(gridGroup)

    for ref in gdfRefDict:
        if gdfRefDict[ref]["geometry"][0].geom_type == 'Point':
            refLayer = folium.GeoJson(data=gdfRefDict[ref]["geometry"],
                marker = folium.CircleMarker(radius = 1, # Radius in metres
                                           weight = 6, #outline weight
                                           color = '#7B7C7E',
                                           fill_opacity = 1))
            refGroup.add_child(refLayer)
        else:
            refStyle = {'color': '#00CED1',"weight": 4,"opacity": 0.65,"fillOpacity": 0,"radius":3}
            refLayer = folium.GeoJson(data=gdfRefDict[ref]["geometry"], style_function = lambda x: refStyle)
            refGroup.add_child(refLayer)

    limStyle = {'color': '#ef4310',"weight": 4,"opacity": 0.65,"fillOpacity": 0,"radius":3}
    limLayer = folium.GeoJson(data=gdfLimit["geometry"], style_function = lambda x: limStyle)
    limGroup.add_child(limLayer)

    meshmap.add_child(limGroup)
    meshmap.add_child(refGroup)

    bPts = gdfMesh.to_crs("EPSG:4326").total_bounds
    xMinyMin = list(bPts[:2])[::-1]
    xMaxyMax = list(bPts[2:])[::-1]
    meshmap.fit_bounds([xMinyMin,xMaxyMax], padding=(1, 1))
    folium.LayerControl('topleft', collapsed= False).add_to(meshmap)

    meshmap = meshmap._repr_html_()

    return meshmap

def pointCloudGraph(self, meshForm, vorMesh, *args, **kwargs):
    project=self.getProject(self.request)
    limitLayer= LimitLayer.objects.get(project=project)
    refinementLayers=RefinementLayer.objects.filter(project=project)

    outNamePointCloud = 'voronoiPointCloud'
    outNameCircleUnion = 'voronoiCircleUnion'

    outNamePointCloudDir = os.path.join(meshForm.getAbsDir,'pointCloudTemp')

    outTempPointCloudPath = os.path.join(outNamePointCloudDir,outNamePointCloud+'.shp')
    outCircleUnionPath = os.path.join(outNamePointCloudDir,'voronoiCircleUnion.shp')
    outVertexBufferPath = os.path.join(outNamePointCloudDir,'vertexBuffer.shp')
    outInteriorsPath = os.path.join(outNamePointCloudDir,'interiors.shp')

    if not os.path.isdir(outNamePointCloudDir):
        os.makedirs(outNamePointCloudDir, exist_ok=True)

    vorMesh.getPointsAsShp('vertexTotal',outTempPointCloudPath)
    vorMesh.getPolyAsShp('circleUnion',outCircleUnionPath)
    vorMesh.getPointsAsShp('vertexBuffer',outVertexBufferPath)
    vorMesh.getPolyAsShp('interiors',outInteriorsPath)

    gdfLimit=read_shp_from_zip(limitLayer.limitFile.path)
    gdfRefDict = {}
    for index, refinemet in enumerate(refinementLayers):
        gdfRefDict[index] = read_shp_from_zip(refinementLayers[index].refinementFile.path)
    gdfMesh=read_shp_from_zip(meshForm.meshFile.path)

    gdfPointCloud=gpd.read_file(outTempPointCloudPath)
    gdfCircleUnion=gpd.read_file(outCircleUnionPath)
    gdfVertexBuffer=gpd.read_file(outVertexBufferPath)
    gdfInteriors=gpd.read_file(outInteriorsPath)

    #add point cloud map
    gdfPointCloud=gdfPointCloud.to_crs("EPSG:4326")
    gdfPtClBds=gdfPointCloud.total_bounds
    ptClMap = folium.Map(max_zoom=20)

    gridGroup = folium.FeatureGroup(name='Grid')
    refGroup = folium.FeatureGroup(name='Refinement Layers')
    limGroup = folium.FeatureGroup(name='Limit Layer')
    pointGroup = folium.FeatureGroup(name='Point Cloud')
    circleUnionGroup = folium.FeatureGroup(name='Circle Union')
    vertexBufferGroup = folium.FeatureGroup(name='Vertex Buffer')
    interiorsGroup = folium.FeatureGroup(name='Interiors')

    pointLayer = folium.GeoJson(data=gdfPointCloud["geometry"],
        marker = folium.CircleMarker(radius = 1, # Radius in metres
                                   weight = 2, #outline weight
                                   fill_color = '#5AAACD',
                                   fill_opacity = 1))
    pointGroup.add_child(pointLayer)

    gridStyle = {'fillColor': '#5AAACD',"weight": 1,"opacity": 0.65,"fillOpacity": 0}
    gridLayer = folium.GeoJson(data=gdfMesh["geometry"], style_function = lambda x: gridStyle)
    gridGroup.add_child(gridLayer)

    for ref in gdfRefDict:
        if gdfRefDict[ref]["geometry"][0].geom_type == 'Point':
            refLayer = folium.GeoJson(data=gdfRefDict[ref]["geometry"],
                marker = folium.CircleMarker(radius = 1, # Radius in metres
                                           weight = 6, #outline weight
                                           color = '#7B7C7E',
                                           fill_opacity = 1))
            refGroup.add_child(refLayer)
        else:
            refStyle = {'color': '#00CED1',"weight": 4,"opacity": 0.65,"fillOpacity": 0,"radius":3}
            refLayer = folium.GeoJson(data=gdfRefDict[ref]["geometry"], style_function = lambda x: refStyle)
            refGroup.add_child(refLayer)

    limStyle = {'color': '#ef4310',"weight": 4,"opacity": 0.65,"fillOpacity": 0,"radius":3}
    limLayer = folium.GeoJson(data=gdfLimit["geometry"], style_function = lambda x: limStyle)
    limGroup.add_child(limLayer)

    circleUnionStyle = {'color': '#80BCD8',"weight": 4,"opacity": 0.65,"fillOpacity": 0}
    circleUnionLayer = folium.GeoJson(data=gdfCircleUnion["geometry"], style_function = lambda x: circleUnionStyle)
    circleUnionGroup.add_child(circleUnionLayer)

    vertexBufferLayer = folium.GeoJson(data=gdfVertexBuffer["geometry"],
        marker = folium.CircleMarker(radius = 4,
                                    weight = .5,
                                   color = '#80BCD8',
                                   fill_opacity = 0.5))
    vertexBufferGroup.add_child(vertexBufferLayer)

    if not gdfInteriors.empty:
        interiorsStyle = {'color': '#B22222',"weight": 4,"opacity": 0.65,"fillOpacity": 0,"radius":3}
        interiorsLayer = folium.GeoJson(data=gdfInteriors["geometry"], style_function = lambda x: interiorsStyle)
        interiorsGroup.add_child(interiorsLayer)

    ptClMap.add_child(gridGroup)
    ptClMap.add_child(limGroup)
    ptClMap.add_child(refGroup)
    ptClMap.add_child(pointGroup)
    ptClMap.add_child(circleUnionGroup)
    ptClMap.add_child(vertexBufferGroup)
    if not gdfInteriors.empty:
        ptClMap.add_child(interiorsGroup)

    bPts = gdfMesh.to_crs("EPSG:4326").total_bounds
    xMinyMin = list(bPts[:2])[::-1]
    xMaxyMax = list(bPts[2:])[::-1]
    ptClMap.fit_bounds([xMinyMin,xMaxyMax], padding=(1, 1))
    folium.LayerControl('topleft', collapsed= False).add_to(ptClMap)
    ptClMap=ptClMap._repr_html_() #up

    return ptClMap

def waterTableGraph(self, c_map, Line, *args, **kwargs):
    project=self.getProject(self.request)
    limitLayer= LimitLayer.objects.get(project=project)
    refinementLayers=RefinementLayer.objects.filter(project=project)
    meshLayer = VoronoiMesh.objects.get(project=project)

    gdfLimit=read_shp_from_zip(limitLayer.limitFile.path)
    gdfRefDict = {}
    for index, refinemet in enumerate(refinementLayers):
        gdfRefDict[index] = read_shp_from_zip(refinementLayers[index].refinementFile.path)
    gdfMesh = read_shp_from_zip(meshLayer.meshFile.path)

    meshmap = folium.Map(max_zoom=20)

    gridGroup = folium.FeatureGroup(name='Grid')
    limGroup = folium.FeatureGroup(name='Limit Layer')
    refGroup = folium.FeatureGroup(name='Refinement Layers')
    contourGroup = folium.FeatureGroup(name='Contours')

    basemap2 = ('https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}')
    folium.WmsTileLayer(
      url=basemap2,
      layers=None,
      name='Google Sattelite',
      attr='Google',
      control=True,
      overlay=False,
    ).add_to(meshmap)

    gridStyle = {'fillColor': '#5AAACD',"weight": 1,"opacity": 0.65,"fillOpacity": 0}
    gridLayer = folium.GeoJson(data=gdfMesh["geometry"], style_function = lambda x: gridStyle)
    gridGroup.add_child(gridLayer)

    meshmap.add_child(gridGroup)

    for ref in gdfRefDict:
        if gdfRefDict[ref]["geometry"][0].geom_type == 'Point':
            refLayer = folium.GeoJson(data=gdfRefDict[ref]["geometry"],
                marker = folium.CircleMarker(radius = 1, # Radius in metres
                                           weight = 6, #outline weight
                                           color = '#7B7C7E',
                                           fill_opacity = 1))
            refGroup.add_child(refLayer)
        else:
            refStyle = {'color': '#00CED1',"weight": 4,"opacity": 0.65,"fillOpacity": 0,"radius":3}
            refLayer = folium.GeoJson(data=gdfRefDict[ref]["geometry"], style_function = lambda x: refStyle)
            refGroup.add_child(refLayer)

    limStyle = {'color': '#ef4310',"weight": 4,"opacity": 0.65,"fillOpacity": 0,"radius":3}
    limLayer = folium.GeoJson(data=gdfLimit["geometry"], style_function = lambda x: limStyle)
    limGroup.add_child(limLayer)

    meshmap.add_child(limGroup)
    meshmap.add_child(refGroup)

    if not c_map.empty:
        red = Color("#97FFFF")
        colors = list(red.range_to(Color("#104E8B"),len(Line)))
        for idx, r in c_map.to_crs("EPSG:4326").iterrows():
             heads_to_map(r,contourGroup,colors[idx])
        meshmap.add_child(contourGroup)

    bPts = gdfMesh.to_crs("EPSG:4326").total_bounds
    xMinyMin = list(bPts[:2])[::-1]
    xMaxyMax = list(bPts[2:])[::-1]
    meshmap.fit_bounds([xMinyMin,xMaxyMax], padding=(1, 1))
    folium.LayerControl('topleft', collapsed= False).add_to(meshmap)

    meshmap = meshmap._repr_html_()

    return meshmap

class Field():
  '''
  Create a Voronoi map that can be used to run Lloyd
  relaxation on an array of 2D points. For background,
  see: https://en.wikipedia.org/wiki/Lloyd%27s_algorithm
  '''

  def __init__(self, *args, **kwargs):
    '''
    Store the points and bounding box of the points to which
    Lloyd relaxation will be applied.
    @param np.array `arr`: a numpy array with shape n, 2, where n
      is the number of 2D points to be moved
    @param float `epsilon`: the delta between the input point
      domain and the pseudo-points used to constrain the points
    '''
    arr = args[0]
    if not isinstance(arr, np.ndarray) or arr.shape[1] != 2:
      raise Exception('Please provide a numpy array with shape n,2')
    self.points = arr
    # find the bounding box of the input data
    self.domains = self.get_domains(arr)
    # ensure no two points have the exact same coords
    self.bb_points = self.get_bb_points(arr)
    self.constrain = kwargs.get('constrain', True)
    self.build_voronoi()

  def constrain_points(self):
    '''
    Update any points that have drifted beyond the boundaries of this space
    '''
    for point in self.points:
      if point[0] < self.domains['x']['min']: point[0] = self.domains['x']['min']
      if point[0] > self.domains['x']['max']: point[0] = self.domains['x']['max']
      if point[1] < self.domains['y']['min']: point[1] = self.domains['y']['min']
      if point[1] > self.domains['y']['max']: point[1] = self.domains['y']['max']


  def get_domains(self, arr):
    '''
    Return an object with the x, y domains of `arr`
    '''
    x = arr[:, 0]
    y = arr[:, 1]
    return {
      'x': {
        'min': min(x),
        'max': max(x),
      },
      'y': {
        'min': min(y),
        'max': max(y),
      }
    }


  def get_bb_points(self, arr):
    '''
    Given an array of 2D points, return the four vertex bounding box
    '''
    return np.array([
      [self.domains['x']['min'], self.domains['y']['min']],
      [self.domains['x']['max'], self.domains['y']['min']],
      [self.domains['x']['min'], self.domains['y']['max']],
      [self.domains['x']['max'], self.domains['y']['max']],
    ])


  def build_voronoi(self):
    '''
    Build a voronoi map from self.points. For background on
    self.voronoi attributes, see: https://docs.scipy.org/doc/scipy/
      reference/generated/scipy.spatial.Voronoi.html
    '''
    # build the voronoi tessellation map
    self.voronoi = Voronoi(self.points, qhull_options='Qbb Qc Qx')

    # constrain voronoi vertices within bounding box
    if self.constrain:
      for idx, vertex in enumerate(self.voronoi.vertices):
        x, y = vertex
        if x < self.domains['x']['min']:
          self.voronoi.vertices[idx][0] = self.domains['x']['min']
        if x > self.domains['x']['max']:
          self.voronoi.vertices[idx][0] = self.domains['x']['max']
        if y < self.domains['y']['min']:
          self.voronoi.vertices[idx][1] = self.domains['y']['min']
        if y > self.domains['y']['max']:
          self.voronoi.vertices[idx][1] = self.domains['y']['max']

  def find_centroid(self, vertices):
    '''
    Find the centroid of a Voroni region described by `vertices`,
    and return a np array with the x and y coords of that centroid.
    The equation for the method used here to find the centroid of a
    2D polygon is given here: https://en.wikipedia.org/wiki/
      Centroid#Of_a_polygon
    @params: np.array `vertices` a numpy array with shape n,2
    @returns np.array a numpy array that defines the x, y coords
      of the centroid described by `vertices`
    '''
    area = 0
    centroid_x = 0
    centroid_y = 0
    for i in range(len(vertices)-1):
      step = (vertices[i  , 0] * vertices[i+1, 1]) - \
             (vertices[i+1, 0] * vertices[i  , 1])
      area += step
      centroid_x += (vertices[i, 0] + vertices[i+1, 0]) * step
      centroid_y += (vertices[i, 1] + vertices[i+1, 1]) * step
    area /= 2
    # prevent division by zero - equation linked above
    if area == 0: area += 0.0000001
    centroid_x = (1.0/(6.0*area)) * centroid_x
    centroid_y = (1.0/(6.0*area)) * centroid_y
    # prevent centroids from escaping bounding box
    if self.constrain:
      if centroid_x < self.domains['x']['min']: centroid_x = self.domains['x']['min']
      if centroid_x > self.domains['x']['max']: centroid_x = self.domains['x']['max']
      if centroid_y < self.domains['y']['min']: centroid_y = self.domains['y']['min']
      if centroid_y > self.domains['y']['max']: centroid_y = self.domains['y']['max']
    return np.array([centroid_x, centroid_y])


  def relax(self):
    '''
    Moves each point to the centroid of its cell in the voronoi
    map to "relax" the points (i.e. jitter the points so as
    to spread them out within the space).
    '''
    centroids = []
    for idx in self.voronoi.point_region:
      # the region is a series of indices into self.voronoi.vertices
      # remove point at infinity, designated by index -1
      region = [i for i in self.voronoi.regions[idx] if i != -1]
      # enclose the polygon
      region = region + [region[0]]
      # get the vertices for this regioncd
      verts = self.voronoi.vertices[region]
      # find the centroid of those vertices
      centroids.append(self.find_centroid(verts))
    self.points = np.array(centroids)
    self.constrain_points()
    self.build_voronoi()


  def get_points(self):
    '''
    Return the input points in the new projected positions
    @returns np.array a numpy array that contains the same number
      of observations in the input points, in identical order
    '''
    return self.points


import geopandas as gpd
import os, shutil, time
from pathlib import Path
import io
import fiona
import numpy as np
from shapely.geometry import Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon, mapping
from shapely.ops import unary_union
import shutil
from collections import OrderedDict


def read_shp_from_zip(file):
    zipshp = io.BytesIO(open(file, 'rb').read())
    with fiona.BytesCollection(zipshp.read()) as src:
        crs = src.crs
        gdf = gpd.GeoDataFrame.from_features(src, crs=crs)
    return gdf

def write_shp_as_zip(zipLoc,zipDest,baseName):
    shutil.make_archive(base_dir=zipLoc,
        root_dir=zipDest,
        format='zip',
        base_name=baseName)

def shpFromZipAsFiona(file):
    zipshp = io.BytesIO(open(file, 'rb').read())
    fionaObj = fiona.BytesCollection(zipshp.read())
    return fionaObj


def remove_files_and_folder(path_to_file, folder=True):
    folder=os.path.dirname(path_to_file)

    if os.path.isfile(path_to_file):
        os.remove(path_to_file)
        print("File has been deleted")
    else:
        print("File does not exist")
    #for filename in os.listdir(folder):
    #    file_path=os.path.join(folder,filename)
        """
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))

    if folder:
        os.rmdir(folder)
    """

def isMultiGeometry(geom):
    return isinstance(geom, (MultiPoint, MultiLineString, MultiPolygon))

#auxiliary funtion to intersect:
def intersectLimitLayer(discLayerGeom, modelDis):
    discGeomList = []  
    #generic 
    if isMultiGeometry(discLayerGeom):
        for partGeom in discLayerGeom.geoms:
            discGeomClip =  modelDis['limitGeometry'].intersection(partGeom)
            if not discGeomClip.is_empty:
                discGeomList.append(discGeomClip)
    else:
        discGeomClip =  modelDis['limitGeometry'].intersection(discLayerGeom)
        if not discGeomClip.is_empty:
            discGeomList.append(discGeomClip)
        else: 
            return False

    unaryGeom = unary_union(discGeomList)

    if isMultiGeometry(unaryGeom):
        unaryFilter = [geom for geom in unaryGeom.geoms]
    else:
        unaryFilter = [unaryGeom]

    return unaryFilter    

def processVertexFilterCloseLimit(layerRef,layerGeom,modelDis,vertexType):
    #first conditional
    if vertexType == 'Org':
        if layerGeom.geom_type == 'Polygon':
            pointObject = layerGeom.exterior.coords.xy
            pointList = list(zip(pointObject[0],pointObject[1]))
        elif layerGeom.geom_type == 'LineString':
            pointObject = layerGeom.coords.xy
            pointList = list(zip(pointObject[0],pointObject[1]))
        elif layerGeom.geom_type == 'Point':
            #covered in the third conditional
            pass
        else:
            print('Something went wrong with the org vertex')
    elif vertexType == 'Dist':
        pointList = []
        if layerGeom.geom_type == 'Polygon':
            polyLength = layerGeom.exterior.length
            pointProg = np.arange(0,polyLength,layerRef)
            for prog in pointProg:
                pointXY = list(layerGeom.exterior.interpolate(prog).xy)
                pointList.append([pointXY[0][0],pointXY[1][0]])
        elif layerGeom.geom_type == 'LineString':
            lineLength = layerGeom.length
            pointProg = np.arange(0,lineLength,layerRef)
            for prog in pointProg:
                pointXY = list(layerGeom.interpolate(prog).xy)
                pointList.append([pointXY[0][0],pointXY[1][0]])
        elif layerGeom.geom_type == 'Point':
            #covered in the third conditional
            pass
        else:
            print('Something went wrong with the dist vertex')

    #second conditional
    if layerGeom.geom_type == 'Polygon' or layerGeom.geom_type == 'LineString':
        if layerGeom.buffer(layerRef).within(modelDis['limitGeometry']):
            filterPointList = pointList
        else:
            filterPointList = []
            for point in pointList:
                pointPoint = Point(point)
                if pointPoint.buffer(layerRef).within(modelDis['limitGeometry']):
                    filterPointList.append(point)
        
        if layerGeom.geom_type == 'Polygon' and len(filterPointList) > 2:
            filterPointListGeom = Polygon(filterPointList)
        elif len(filterPointList) > 1:
            filterPointListGeom = LineString(filterPointList)
        elif len(filterPointList) == 1:
            filterPointListGeom = Point(filterPointList)
        else:
            filterPointListGeom = None

    #third conditional
    elif layerGeom.geom_type == 'Point':
        pointObject = layerGeom.coords.xy
        point = (pointObject[0][0],pointObject[1][0])
        if layerGeom.buffer(layerRef).within(modelDis['limitGeometry']):
            filterPointList = [point]
            filterPointListGeom = layerGeom
        else:
            filterPointList = []
            filterPointListGeom = None

    return filterPointList, filterPointListGeom 

def getFionaDictPoly(polyGeom, index):
    polyCoordList = []
    x,y = polyGeom.exterior.coords.xy
    polyCoordList.append(list(zip(x,y)))
    if polyGeom.interiors[:] != []:
        interiorList = []
        for interior in polyGeom.interiors:
            polyCoordList.append(interior.coords[:])
    feature = {
        "geometry": {'type':'Polygon',
                    'coordinates':polyCoordList},
        "properties": OrderedDict([("id",index)]),
    }
    return feature

def getFionaDictPoint(pointGeom, index):
    if isinstance(pointGeom[0], float):
        feature = {
                "geometry": {'type':'Point',
                            'coordinates':(pointGeom[0],pointGeom[1])},
                "properties": OrderedDict([("id",index)]),
            }
        return feature

def initiateOutputFolder(outputPath):
    if os.path.isdir(outputPath):
        print('The output folder %s exists and has been cleared'%outputPath)
        shutil.rmtree(outputPath)
        os.mkdir(outputPath)
    else:
        os.mkdir(outputPath)
        print('The output folder %s has been generated.'%outputPath)

###############
# Output functions
###############

def getVoronoiAsShp(modelDis, shapePath=''):
    print('\n/----Generation of the voronoi shapefile----/')
    start = time.time()
    schema_props = OrderedDict([("id", "int")])
    schema={"geometry": "Polygon", "properties": schema_props}

    outFile = fiona.open(shapePath,mode = 'w',driver = 'ESRI Shapefile',
                        crs = modelDis['crs'], schema=schema)
    
    for index, poly in enumerate(modelDis['voronoiRegions'].geoms):
        polyCoordList = []
        x,y = poly.exterior.coords.xy
        polyCoordList.append(list(zip(x,y)))
        if poly.interiors[:] != []:
            interiorList = []
            for interior in poly.interiors:
                polyCoordList.append(interior.coords[:])
        feature = {
            "geometry": {'type':'Polygon',
                        'coordinates':polyCoordList},
            "properties": OrderedDict([("id",index)]),
        }
        outFile.write(feature)
    outFile.close()

    
    end = time.time()
    print('\nTime required for voronoi shapefile: %.2f seconds \n'%(end - start), flush=True)

def getPolyAsShp(modelDis,circleList,shapePath=''):
    start = time.time()
    schema_props = OrderedDict([("id", "str")])
    schema={"geometry": "Polygon", "properties": schema_props}
    
    outFile = fiona.open(shapePath,mode = 'w',driver = 'ESRI Shapefile',
                        crs = modelDis['crs'], schema=schema)
    
    if isinstance(modelDis[circleList], dict):
        for key, value in modelDis[circleList].items():
            if isMultiGeometry(value):
                for index, poly in enumerate(value.geoms):
                    feature = getFionaDictPoly(poly, index)
                    outFile.write(feature)

    if isMultiGeometry(modelDis[circleList]):
        for index, poly in enumerate(modelDis[circleList].geoms):
            feature = getFionaDictPoly(poly, index)
            outFile.write(feature)
    else:
        poly = modelDis[circleList]
        feature = getFionaDictPoly(poly, '1')
        outFile.write(feature)
    outFile.close()
    
    end = time.time()
    print('\nTime required for voronoi shapefile: %.2f seconds \n'%(end - start), flush=True)

def getPointsAsShp(modelDis,pointList,shapePath=''):
    schema_props = OrderedDict([("id", "str")])
    schema={"geometry": "Point", "properties": schema_props}
    if shapePath != '':
        outFile = fiona.open(shapePath,mode = 'w',driver = 'ESRI Shapefile',
                            crs = modelDis['crs'], schema=schema)
        if isinstance(modelDis[pointList], dict):
            #print(modelDis[pointList].keys())
            for key, value in modelDis[pointList].items():
                for index, point in enumerate(value):
                    feature = getFionaDictPoint(point, index)
                    if feature != None:
                        outFile.write(feature)
                    else:
                        print('Something went wrong with %s'%point)
        else:
            for index, point in enumerate(modelDis[pointList]):
                feature = getFionaDictPoint(point, index)
                if feature != None:
                    outFile.write(feature)
                else:
                    print('Something went wrong with %s'%point)
        outFile.close()

#########
# miscelaneous functions
#########

def printHeader():
    print('''
                                                                                                    
*mSi                                                                                       
gQQ>                                                                                       
dQU;                                 +|:                                     :v)_          
;PQm'                                %B$s                                    .gQMe          
PYQ7.                               -3QE_                                     <e}'          
c8Qx                                '$RT                                                    
?HM"   )7yw1=       .)r]jJfzi.   `=>!QDuvxxi_   `<s[LCwe<    ,>seua:  ^!C3o' `eur           
oRk= vdZ6qDQE"     ]PF)/+vJBNe`  :l{6Q8!I![s' .ebJ<//%3MD]   )fffQDv.ebPhZY/ QQQ#           
JQS'7b]_  ?Q$r     EWy     3Qg^     pQX       :GWj    _mQ5'     lQ&TT4v   .  rQDl           
5QnCV/    ]QZ<      :"/iiss4Q5:    -GQS        .:|/<v%IPQJ`     !Qk[h;       ?QK"           
.SQXd^    _JQh:    -*53e*ppaDQL.    _&QF       '!6fa{vzzMQa      7Q@q|        1Q@,           
;4QWi     :pQp    _dQY-   xm@Qa     _OQd      ^XQV.   rhHQ!     .wQBo         [QK^           
KkQ6'     '2QO}vc/:&QDa}75L<2Qgx="= .nQMCv)%I"UUQ81}j57>SQhi=|; .dQA'         %QQCi%)_       
/jJ>       82mw[i: /zmVFa|  ;t53j}+  `1mVpn!>_ UuSh21/  =oFy7{; .z#I          .nm57r/.       
                                                                                                                                                                                                                                      
''')
    
def copyTemplate(templateType, prefix = ''):
    utilsPath = os.path.realpath(__file__)
    examplePath = os.path.join(os.path.dirname(utilsPath),'examples','notebooks')
    workingPath = os.getcwd()

    def generateSrcDstPath():
        srcPath = str(Path(os.path.join(examplePath, templateName)))
        if prefix != '':
            dstPath = str(Path(os.path.join(workingPath, prefix+'_'+templateName)))
        else:
            dstPath = str(Path(os.path.join(workingPath, templateName)))
        return srcPath,dstPath

    if templateType == 'generateVoronoi':
        templateName = 'p1_'+templateType+'.ipynb'
        srcPath,dstPath = generateSrcDstPath()
        shutil.copy2(srcPath,dstPath)
    elif templateType == 'modelCreation':
        templateName = 'p2_'+templateType+'.ipynb'
        srcPath,dstPath = generateSrcDstPath()
        shutil.copy2(srcPath,dstPath)
    elif templateType == 'obsCalculated':
        templateName = 'p3_'+templateType+'.ipynb'
        srcPath,dstPath = generateSrcDstPath()
        shutil.copy2(srcPath,dstPath)
    elif templateType == 'vtkGeneration':
        templateName = 'p4_'+templateType+'.ipynb'
        srcPath,dstPath = generateSrcDstPath()
        shutil.copy2(srcPath,dstPath)
    else:
        print("The template you want doesn't exists capullo")
